/*

 compat.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "compat.h"
#include "driver.h"
#include "ideserv.h"
#include "path.h"
#include "crc8.h"
#include "partition.h"
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define OPEN_RONLY 0
#define OPEN_STATUS 2
#define OPEN_WONLY 4
#define OPEN_ERR 6

int openfile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    c64c_t cmd[256];
    unsigned char status=OPEN_ERR;
    unsigned char mode='R';
    unsigned char channel,fp;
    unsigned long long l=0;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];

    fp=0;
    while (fp<255) {
        r=io_ops->getb();
        if (r < 0) {
            if (arguments->mode == M_ETHERNET) break;
            return 1;
        }
        cmd[fp]=r;
        if (!cmd[fp]) break;
        fp++;
    }
    cmd[fp]=0;

    if (arguments->mode != M_ETHERNET && usecrc) {
        int fr;
        if (io_ops->getb() < 0) return 1;
        if (crc_get()) {
	    seterror(42, 1);
            fprintf(lf, "CRC error while open\n");
            io_ops->getb();
            return 1;
        }
        fr = io_ops->getb();
        if (fr != 0x00) {
	    seterror(41, 1);
            if (fr > 0) fprintf(lf, "Frame error %02x while open\n", fr);
            return 1;
        }
    }

    alarm(0);

    if (arguments->debug) {
        int i=0;
        fprintf(lf, "Command:\n");
        while (cmd[i]) {
            fprintf(lf, " %02x", cmd[i++]);
            if (!(i & 15)) putc('\n', lf);
        }
        if (i & 15) putc('\n', lf);
    }
    if (buffer->file) {
        fclose(buffer->file);
        buffer->file=NULL;
    }
    buffer->mode = CM_CLOSED;

    if (channel == 15) {
	buffer->mode = CM_ERR;
        fprintf(lf,"Open [%d]: \"%s\": ",channel,cmd);
	commandchannel(cmd);
	status=OPEN_STATUS;
    } else if (cmd[0]=='$') {
        unsigned char dirline[32];
        struct dirent *ep;
        DIR *dp;
        int sum=0,partition;
        char outpath[1000];
        buffer->mode=CM_DIR;
        buffer->length=buffer->pointer=0;
        c64c_t name[17], type[4]={'*',0};
        const c64c_t *outname;
        partition_t outpart = 0;

        partition=!strncmp((char *)&cmd[1],"=P",2);

        resolv_path(cmd+1, outpath, &outname, &outpart, lf, arguments->relaxfiletypes);
        fprintf(lf,"Open [%d]: \"$%s\"\n", channel, outpath);
        convertfilename(outname, '=', name, type, NULL);

        if (!name[0]) strcpy((char *)name, "*");

        if (channel==1) {
            seterror(70,0); status=OPEN_ERR;
        } else if (channel>1) {
            dp=opendir(outpath[0]?outpath:".");
            if (!dp) {
                errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                status=OPEN_ERR;
                goto vege;
            }
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            buffer->data[buffer->length++]='I';

            memset(&dirline[0],32,16);
            memset(&dirline[16],0,8);
            dirline[24]=A_DIR;
            memcpy(&dirline[25],"DIR",3);
            memset(&dirline[28],0,4);
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                char lname2[2020];

                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/",999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, (DIRECTORY_ENTRY_T*)dirline, arguments->relaxfiletypes)) continue;

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            }
            initdirentry(1);

            errtochannel15(closedir(dp));
            status=OPEN_RONLY;
        } else if (partition) {
            int i, n=0;
            memcpy(&dirline[0x00],"\1\4\1\1\377\0\22\42",8);
            memcpy(&dirline[0x08],"IDESERVD " VERSION2 "  ",16);
            memcpy(&dirline[0x18],"\42\40" SYSTEMNAME "\0",8);
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            for (i=1;i<255;i++) {
                c64c_t *name = partition_get_name(i);
                if (!name) continue;
                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=i;
                dirline[0x03]=0;
                memcpy(&dirline[0x04],"   \42\42                 NET  \0",28);
                memcpy(&dirline[0x08], name, strlen((char *)name));
                dirline[0x08 + strlen((char *)name)] = '"';
                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                n++;
            }
            dirline[0x00]=dirline[0x01]=1;
            dirline[0x02]=n;
            dirline[0x03]=0;
            memcpy(&dirline[0x04],"PARTITIONS.              \0\0\0",28);
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            errtochannel15(0);
            status=OPEN_RONLY;
        } else {

            dp=opendir(outpath[0]?outpath:".");

            if (!dp) {
                errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
                status = OPEN_ERR;
                goto vege;
            }

            memcpy(&dirline[0x00],"\1\4\1\1\1\0\22\42",8);
            dirline[0x04] = outpart ? outpart : partition_get_current();
            memset(&dirline[0x08],32,16);
            memcpy(&dirline[0x18],"\42\40" SYSTEMNAME "\0",8);
            if (!buffer->data) {
                buffer->data=(unsigned char *)malloc(65536);
                if (!buffer->data) exit(1);//out of memory?!
            }
            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];

            initdirentry(0);
            while ((ep=readdir(dp)))
            {
                DIRECTORY_ENTRY_T dirent;
                int offs;
                unsigned int n;
                char lname2[2020];

                strcpy(lname2,outpath);
                if (outpath[0]) strncat(lname2,"/",999);
                strncat(lname2,ep->d_name,999);

                if (createdirentry(ep->d_name, lname2, &dirent, arguments->relaxfiletypes)) continue;

                if (dirent.pointer[0] & 0x80) continue;
                n=(dirent.filesize[0]+(dirent.filesize[1] << 8)+(dirent.filesize[2] << 16)+(dirent.filesize[3] << 24)+255) >> 8;
                if (n>65535) n=65535;

                dirent.filesize[0]=0;
                if (!matchname(dirent.name, name)) continue;
                dirent.time[0]=0;
                if (!matchname(dirent.filetype,type)) continue;

                dirline[0x00]=dirline[0x01]=1;
                dirline[0x02]=n;
                dirline[0x03]=n >> 8;
                sum+=n;
                memset(&dirline[0x04],32,27);
                offs=7;
                if (n>9) offs--;
                if (n>99) offs--;
                if (n>999) offs--;
                dirline[offs++]=0x22;
                n=0;while (dirent.name[n] && n<16) dirline[offs++]=dirent.name[n++];
                dirline[offs++]=0x22;
                offs+=(16-n);

                dirline[offs++]=(dirent.attrib & A_CLOSED)?' ':'*';

                n=0;while (n<3 && dirent.filetype[n]) {dirline[offs+n]=dirent.filetype[n];n++;}
                if ((dirent.attrib & (A_DELETEABLE | A_WRITEABLE))!=(A_DELETEABLE | A_WRITEABLE)) dirline[offs+3]='<';

                for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
                if (sum>65535) sum=65535;
            }
            initdirentry(1);

            dirline[0x00]=dirline[0x01]=1;
            dirline[0x02]=sum;
            dirline[0x03]=sum >> 8;
            memcpy(&dirline[0x04],"BLOCKS USED.             \0\0\0",28);

            for (l=0;l<32;l++) buffer->data[buffer->length++]=dirline[l];
            errtochannel15(closedir(dp));
            status=OPEN_RONLY;
        }
    } else {
        c64c_t name[17], type[4]={'*',0};
        const c64c_t *outname;
        unsigned char overwrite;
        char lname[1000],outpath[1020];
        struct dirent *ep;
        DIR *dp;
        int found=0;
        int b;

        overwrite=0;
        for (b=0;cmd[b];b++) if (cmd[b]==':') {
            if (cmd[0]=='@') overwrite=1;
            break;
        }

        if (!cmd[0]) goto vege;

        resolv_path(cmd, outpath, &outname, NULL, lf, arguments->relaxfiletypes);
        if (channel==1) strcpy((char *)type,"PRG");
        convertfilename(outname, ',', name, type, &mode);

        dp=opendir(outpath[0]?outpath:".");
        if (!dp) {
            errtochannel15(1);fprintf(lf,"Couldn't open the directory \"%s\".\n",outpath);
            status = OPEN_ERR;
            goto vege;
        }

        initdirentry(0);
        while ((ep=readdir(dp)))
        {
            DIRECTORY_ENTRY_T dirent;
            char lname2[2020];

            strncpy(lname,ep->d_name,999);
            strcpy(lname2,outpath);
            if (outpath[0]) strncat(lname2,"/", 999);
            strncat(lname2,ep->d_name,999);

            if (createdirentry(ep->d_name, lname2, &dirent, arguments->relaxfiletypes)) continue;

            if ((dirent.attrib & (A_ANY | A_CLOSED))!=(A_CLOSED | A_NORMAL)) continue;

            dirent.filesize[0]=0;
            if (!matchname(dirent.name, name)) continue;
            dirent.time[0]=0;
            if (!matchname(dirent.filetype,type)) continue;

            found=1;
            break;
        }
        closedir(dp);
        initdirentry(1);
        status=OPEN_ERR;

        if (!found) {
            mbstate_t ps;
            int f=0,f2=0,g=0;
            memset(&ps, 0, sizeof(ps));
            while (name[f] && f < 16) {
                f2 += c64toascii(lname + f2, name[f], &ps);
                f++;
            }
            if (!arguments->relaxfiletypes || memcmp(type,"PRG",3)) {
                memset(&ps, 0, sizeof(ps));
                lname[f2++]=',';
                while (type[g] && g < 3) {
                    f2 += c64toascii(lname + f2, type[g], &ps);
                    g++;
                }
            }
            lname[f2]=0;
        }

        if (channel==0) mode='R';
        if (channel==1) mode='W';

        if (outpath[0]) strcat(outpath,"/");
        strcat(outpath,lname);

        fprintf(lf,"Open [%d]: \"%s\" %c\n",channel,outpath,mode);

        if (!name[0]) {
            seterror(34,0);goto vege;
        }

        if (mode=='W') {
            int f=0;

            while (lname[f]) {
                if (lname[f]=='*' || lname[f]=='?' || lname[f]==':' || lname[f]=='=') {
                    seterror(33,0);goto vege;
                }
                f++;
            }

            if (found && !overwrite) {
                seterror(63,0);
            } else {
#ifndef WIN32
                buffer->fd=open(outpath, O_CREAT|O_WRONLY|(overwrite ? O_TRUNC : O_EXCL), S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
#else
                buffer->fd=open(outpath, O_CREAT|O_WRONLY|(overwrite ? O_TRUNC : O_EXCL)|O_BINARY, S_IRUSR|S_IWUSR);
#endif
                if (buffer->fd<0) errtochannel15(1); else {
                    status=OPEN_WONLY;
                    buffer->file=fdopen(buffer->fd,"wb");
                    buffer->mode=CM_COMPAT;
                }
            }
        } else {
            if (!found) {
                seterror(62,0);
            } else {
                if (mode=='R') {
                    buffer->fd=open(outpath, O_RDONLY|O_BINARY, 0);
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=OPEN_RONLY;//ok
                        buffer->file=fdopen(buffer->fd, "rb");
                        buffer->mode=CM_COMPAT;
                    }
                }
                if (mode=='A') {
                    buffer->fd=open(outpath, O_WRONLY|O_BINARY, 0);
                    if (buffer->fd<0) errtochannel15(1); else {
                        status=OPEN_WONLY;//ok
                        buffer->file=fdopen(buffer->fd, "ab");
                        buffer->mode=CM_COMPAT;
                    }
                }
            }
        }
    }
vege:
    if (io_ops->sendb(status)) return 1;
    if (arguments->mode == M_ETHERNET && io_ops->sendb(status)) return 1;
    if (io_ops->flush()) return 1;
    return 0;
}

int readfile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel,status=0;
    unsigned short int bytes,i,j;
    struct timeval start, end;
    unsigned long long l=0,l2;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];
    r=io_ops->getb();if (r < 0) return 1;
    bytes=r;
    if (arguments->mode == M_ETHERNET) {
        if (!bytes) bytes=256;
    } else {
        int fr;
        r=io_ops->getb();if (r < 0) return 1;
        bytes|=r << 8;
        if (usecrc) {
            if (io_ops->getb() < 0) return 1;
            if (crc_get()) {
                seterror(42, 1);
                fprintf(lf, "CRC error while read\n");
                io_ops->getb();
                return 1;
            }
        }
        fr = io_ops->getb();
        if (fr != 0x00) {
	    seterror(41, 1);
            if (fr > 0) fprintf(lf, "Frame error %02x while read\n", fr);
            return 1;
        }
    }

    alarm(0);

    fprintf(lf,"Read [%d] %d bytes",channel,bytes);
    if (arguments->mode != M_ETHERNET) {
        if (io_ops->sendb(0)) return 1;
    }
    if (buffer->mode!=CM_DIR && buffer->mode!=CM_ERR && (buffer->mode!=CM_COMPAT || !buffer->file)) {
        fprintf(lf, (buffer->mode == CM_CLOSED) ? " channel not open\n" : " not readable.\n");
        crc_clear();
        if (io_ops->sendb(0)) return 1;
        if (io_ops->sendb(0)) return 1;
        if (arguments->mode != M_ETHERNET) {
            if (usecrc) {
                if (io_ops->sendb(crc_get())) return 1;
            }
            if (io_ops->sendb(0)) return 1;
        }
        if (io_ops->flush()) return 1;
        return 0;
    }
    if (buffer->mode==CM_COMPAT) {
        unsigned long itt;

        if (!buffer->data) {
            buffer->data=(unsigned char *)malloc(65536);
            if (!buffer->data) return 1;//out of memory?!
        }

        l=bytes=fread(buffer->data,1,bytes?bytes:65536,buffer->file);
        if (feof(buffer->file)) status=0x42; else
        {
            itt=ftell(buffer->file);
            fgetc(buffer->file);
            if (feof(buffer->file)) status=0x40;
            fseek(buffer->file,itt,SEEK_SET);
        }
        j=0;
    } else {
        unsigned int length = buffer->length;
        if (buffer->mode==CM_ERR) {
            buffer->data[length++] = '\r';
        }
        j=buffer->pointer;
        if (bytes==(length-j)) status=0x40;
        if (bytes>length-j) {bytes=length-j;status=0x42;}
        l=bytes;
        buffer->pointer+=bytes;
    }

    if (arguments->mode == M_ETHERNET) {
        if (io_ops->sendb(bytes>0)) return 1;
        if (io_ops->sendb(bytes)) return 1;
        for (i=0;i<bytes;i++) io_ops->sendb(buffer->data[bytes-i-1+j]);
        if (bytes & 1) if (io_ops->sendb(status)) return 1;
        if (io_ops->sendb(status)) return 1;
        if (io_ops->sendb(status)) return 1;
        if (io_ops->flush()) return 1;
    } else {
        crc_clear();
        gettimeofday(&start,NULL);
        if (io_ops->sendb(~bytes)) return 1;
        if (io_ops->sendb(~bytes >> 8)) return 1;
        if (bytes) {
            if (io_ops->sendbytes(bytes, buffer->data+j)) return 1;
        } else {
            if (io_ops->sendbytes(32768, buffer->data)) return 1;
            if (io_ops->sendbytes(32768, buffer->data+32768)) return 1;
        }
	if (usecrc) {
	    if (io_ops->sendb(crc_get())) return 1;
	}
        if (io_ops->sendb(status)) return 1;//fprintf(lf,"Status: %02X\n",status);
        if (io_ops->flush()) return 1;
        gettimeofday(&end,NULL);
        l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
        fprintf(lf,", sent %d bytes",bytes);
        if (l2) {
            l=l*1000000/l2;
            fprintf(lf," at %lld bytes/seconds",l);
        }
    }
    putc('\n', lf);
    if (buffer->mode==CM_ERR && buffer->length + 1 >= buffer->pointer) seterror(0,0);
    return 0;
}

int closefile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];
    if (arguments->mode != M_ETHERNET) {
        int fr;
        if (usecrc) {
            if (io_ops->getb() < 0) return 1;
            if (crc_get()) {
                seterror(42, 1);
                fprintf(lf, "CRC error while close\n");
                io_ops->getb();
                return 1;
            }
        }
        fr = io_ops->getb();
        if (fr != 0x00) {
	    seterror(41, 1);
            if (fr > 0) fprintf(lf, "Frame error %02x while close\n", fr);
            return 1;
        }
    }

    alarm(0);

    fprintf(lf,"Close [%d]\n",channel);
    if (buffer->mode==CM_COMPAT) {
        if (buffer->file) {
            fclose(buffer->file);
            buffer->file=NULL;
        }
        if (channel == 15) buffer->mode = CM_ERR;
        else {
            buffer->mode = CM_CLOSED;
            if (buffer->data) {
                free(buffer->data);
                buffer->data=NULL;
            }
        }
    }

    if (io_ops->sendb(0)) return 1;
    if (arguments->mode == M_ETHERNET && io_ops->sendb(0)) return 1;
    if (io_ops->flush()) return 1;
    return 0;
}

int writefile_compat(const struct io_ops_s *io_ops, struct buffer_s *buff, FILE *lf, const struct arguments_s *arguments, int usecrc) {
    unsigned char channel;
    unsigned short int bytes;
    struct timeval start,end;
    unsigned long long l=0,l2;
    struct buffer_s *buffer;
    int r;

    r=io_ops->getb();if (r < 0) return 1;
    channel=r & 0x0f;
    buffer=&buff[channel];
    r=io_ops->getb();if (r < 0) return 1;
    bytes=r;
    if (arguments->mode == M_ETHERNET) {
        if (io_ops->getb() < 0) return 1;
    } else {
        int fr;
        r=io_ops->getb();if (r < 0) return 1;
        bytes|=r << 8;
        if (usecrc) {
            if (io_ops->getb() < 0) return 1;
            if (crc_get()) {
                seterror(42, 1);
                fprintf(lf, "CRC error while write\n");
                io_ops->getb();
                return 1;
            }
        }
        fr = io_ops->getb();
        if (fr != 0x00) {
	    seterror(41, 1);
            if (fr > 0) fprintf(lf, "Frame error %02x while write\n", fr);
            return 1;
        }
    }

    alarm(0);

    fprintf(lf,"Write [%d] %d bytes",channel,bytes);
    if (buffer->mode==CM_ERR) {
        unsigned char cmd[256];
        int fr;

        if (arguments->mode == M_ETHERNET) {
            if (io_ops->getbytes(bytes, cmd)) return 1;
            if (io_ops->sendb(0)) return 1;
            if (io_ops->sendb(0)) return 1;
        } else {
            if (io_ops->sendb(0)) return 1;
            if (io_ops->flush()) return 1;
            crc_clear();
            if (io_ops->getbytes(bytes, cmd)) return 1;

            if (usecrc) {
                if (io_ops->getb() < 0) return 1;
                if (crc_get()) {
                    seterror(42, 1);
                    fprintf(lf, " CRC error\n");
                    io_ops->getb();
                    return 1;
                }
            }
            fr = io_ops->getb();
            if (fr != 0x00) {
                seterror(41, 1);
                if (fr > 0) fprintf(lf, " frame error %02x\n", fr);
                return 1;
            }

            if (io_ops->sendb(0)) return 1;
        }
        if (io_ops->flush()) return 1;
        cmd[bytes-(bytes && cmd[bytes-1]==0x0d)]=0;
        putc('\n', lf);
        commandchannel(cmd);
        return 0;
    } else {
        if (buffer->mode!=CM_COMPAT) {
            fprintf(lf," not writeable.\n");
            if (io_ops->sendb(2)) return 1;
            if (arguments->mode == M_ETHERNET && io_ops->sendb(2)) return 1;
            if (io_ops->flush()) return 1;
            return 1;
        }

        if (!buffer->data) {
            buffer->data=(unsigned char *)malloc(65536);
            if (!buffer->data) return 1;//out of memory?!
        }

        if (arguments->mode == M_ETHERNET) {
            if (io_ops->getbytes(bytes, buffer->data)) return 1;
            if (fwrite(buffer->data,1,bytes,buffer->file)==bytes) {
                if (io_ops->sendb(0)) return 1;
                if (io_ops->sendb(0)) return 1;
            } else {
                if (io_ops->sendb(2)) return 1;
                if (io_ops->sendb(2)) return 1;
            }
            if (io_ops->flush()) return 1;
        } else {
            int fr;
            if (io_ops->sendb(0)) return 1;
            if (io_ops->flush()) return 1;
            crc_clear();

            gettimeofday(&start,NULL);
            if (io_ops->getbytes(bytes, buffer->data)) return 1;
            gettimeofday(&end,NULL);
	
            if (usecrc) {
                if (io_ops->getb() < 0) return 1;
                if (crc_get()) {
                    seterror(42, 1);
                    fprintf(lf, " CRC error\n");
                    io_ops->getb();
                    return 1;
                }
            }
            fr = io_ops->getb();
            if (fr != 0x00) {
                seterror(41, 1);
                if (fr > 0) fprintf(lf, " frame error %02x\n", fr);
                return 1;
            }

            if (fwrite(buffer->data,1,bytes,buffer->file)==bytes) {
                if (io_ops->sendb(0)) return 1;
            } else { 
                if (io_ops->sendb(2)) return 1;
            }
            if (io_ops->flush()) return 1;
            l=bytes;
            l2=((end.tv_sec-start.tv_sec)*1000000+end.tv_usec-start.tv_usec);
            if (l2) {
                l=l*1000000/l2;
                fprintf(lf," at %lld bytes/seconds",l);
            }
        }
    }
    putc('\n', lf);
    return 0;
}

